# See "Writing benchmarks" in the asv docs for more information.
# https://asv.readthedocs.io/en/latest/writing_benchmarks.html
# or the napari documentation on benchmarking
# https://github.com/napari/napari/blob/main/docs/BENCHMARKS.md
import os

import numpy as np
from packaging.version import parse as parse_version

import napari
from napari.components import Dims
from napari.layers import Points

from .utils import Skip

NAPARI_0_4_19 = parse_version(napari.__version__) <= parse_version('0.4.19')


class Points2DSuite:
    """Benchmarks for the Points layer with 2D data"""

    params = [2**i for i in range(4, 18, 2)]

    if 'PR' in os.environ:
        skip_params = [(2**i,) for i in range(8, 18, 2)]

    def setup(self, n):
        np.random.seed(0)
        self.data = np.random.random((n, 2))
        self.layer = Points(self.data)

    def time_create_layer(self, n):
        """Time to create layer."""
        Points(self.data)

    def time_refresh(self, n):
        """Time to refresh view."""
        self.layer.refresh()

    def time_set_view_slice(self, n):
        """Time to set view slice."""
        if hasattr(self.layer, '_slicing_state'):
            self.layer._slicing_state._set_view_slice()
        else:
            # before https://github.com/napari/napari/pull/8254
            self.layer._set_view_slice()

    def time_update_thumbnail(self, n):
        """Time to update thumbnail."""
        self.layer._update_thumbnail()

    def time_get_value(self, n):
        """Time to get current value."""
        self.layer.get_value((0,) * 2)

    def time_add(self, n):
        self.layer.add(self.data)

    def mem_layer(self, n):
        """Memory used by layer."""
        return self.layer

    def mem_data(self, n):
        """Memory used by raw data."""
        return self.data


class Points3DSuite:
    """Benchmarks for the Points layer with 3D data."""

    params = [2**i for i in range(4, 18, 2)]
    if 'PR' in os.environ:
        skip_params = [(2**i,) for i in range(6, 18, 2)]

    def setup(self, n):
        np.random.seed(0)
        self.data = np.random.random((n, 3))
        self.layer = Points(self.data)

    def time_create_layer(self, n):
        """Time to create layer."""
        Points(self.data)

    def time_refresh(self, n):
        """Time to refresh view."""
        self.layer.refresh()

    def time_set_view_slice(self, n):
        """Time to set view slice."""
        if hasattr(self.layer, '_slicing_state'):
            self.layer._slicing_state._set_view_slice()
        else:
            # before https://github.com/napari/napari/pull/8254
            self.layer._set_view_slice()

    def time_update_thumbnail(self, n):
        """Time to update thumbnail."""
        self.layer._update_thumbnail()

    def time_get_value(self, n):
        """Time to get current value."""
        self.layer.get_value((0,) * 3)

    def mem_layer(self, n):
        """Memory used by layer."""
        return self.layer

    def mem_data(self, n):
        """Memory used by raw data."""
        return self.data


class PointsSlicingSuite:
    """Benchmarks for slicing the Points layer with 3D data."""

    params = [True, False]
    timeout = 300
    skip_params = Skip(always=lambda _: NAPARI_0_4_19)

    def setup(self, flatten_slice_axis):
        np.random.seed(0)
        size = 20000 if 'PR' in os.environ else 20000000
        self.data = np.random.uniform(size=(size, 3), low=0, high=500)
        if flatten_slice_axis:
            self.data[:, 0] = np.round(self.data[:, 0])
        self.layer = Points(self.data)
        self.dims = Dims(ndim=3, point=(249, 0, 0))

    def time_slice_points(self, flatten_slice_axis):
        """Time to take one slice of points"""
        if hasattr(self.layer, '_slicing_state'):
            self.layer._slicing_state._make_slice_request(self.dims)()
        else:
            # before https://github.com/napari/napari/pull/8254
            self.layer._make_slice_request(self.dims)()


class PointsToMaskSuite:
    """Benchmarks for creating a binary image mask from points."""

    param_names = ['num_points', 'mask_shape', 'point_size']
    params = [
        [64, 256, 1024, 4096, 16384],
        [
            (256, 256),
            (512, 512),
            (1024, 1024),
            (2048, 2048),
            (128, 128, 128),
            (256, 256, 256),
            (512, 512, 512),
        ],
        [5, 10],
    ]

    skip_params = Skip(
        if_in_pr=lambda num_points, mask_shape, points_size: num_points > 256
        or mask_shape[0] > 512
    )

    def setup(self, num_points, mask_shape, point_size):
        np.random.seed(0)
        data = np.random.random((num_points, len(mask_shape))) * mask_shape
        self.layer = Points(data, size=point_size)

    def time_to_mask(self, num_points, mask_shape, point_size):
        self.layer.to_mask(shape=mask_shape)


if __name__ == '__main__':
    from utils import run_benchmark

    run_benchmark()



from qtpy.QtWidgets import (
    QCheckBox,
    QWidget,
)

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import checked_to_bool, qt_signals_blocked
from napari.layers.base.base import Layer
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtOutSliceCheckBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer out of slice
    display attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Layer
        An instance of a napari layer.

    Attributes
    ----------
    out_of_slice_checkbox : qtpy.QtWidgets.QCheckBox
        Checkbox to indicate whether to render out of slice.
    out_of_slice_checkbox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the out of slice display enablement chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Layer) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.out_of_slice_display.connect(
            self._on_out_of_slice_display_change
        )

        # Setup widgets
        self.out_of_slice_checkbox = QCheckBox()
        self.out_of_slice_checkbox.setToolTip(trans._('Out of slice display'))
        self.out_of_slice_checkbox.setChecked(self._layer.out_of_slice_display)
        connect_setattr(
            self.out_of_slice_checkbox.stateChanged,
            layer,
            'out_of_slice_display',
            convert_fun=checked_to_bool,
        )

        self.out_of_slice_checkbox_label = QtWrappedLabel(
            trans._('out of slice:')
        )

    def _on_out_of_slice_display_change(self) -> None:
        """Receive layer model out_of_slice_display change event and update checkbox."""
        with qt_signals_blocked(self.out_of_slice_checkbox):
            self.out_of_slice_checkbox.setChecked(
                self._layer.out_of_slice_display
            )

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.out_of_slice_checkbox_label, self.out_of_slice_checkbox)]



from collections.abc import Callable

import numpy as np
import numpy.typing as npt

from napari.layers._scalar_field._slice import _ScalarFieldSliceRequest
from napari.layers.image._image_constants import ImageProjectionMode
from napari.types import ArrayLike


class _ImageSliceRequest(_ScalarFieldSliceRequest):
    @staticmethod
    def _project_slice(
        data: ArrayLike, axis: tuple[int, ...], mode: ImageProjectionMode
    ) -> npt.NDArray:
        """Project a thick slice along axis based on mode."""
        if all(data.shape[axis] == 1 for axis in axis):
            # If all axes are of size 1, return the data as is
            return data[
                tuple(
                    slice(None) if i not in axis else 0
                    for i in range(data.ndim)
                )
            ]
        func: Callable
        if mode == ImageProjectionMode.SUM:
            func = np.sum
        elif mode == ImageProjectionMode.MEAN:
            func = np.mean
        elif mode == ImageProjectionMode.MAX:
            func = np.max
        elif mode == ImageProjectionMode.MIN:
            func = np.min
        else:
            raise NotImplementedError(f'unimplemented projection: {mode}')
        return func(data, tuple(axis))



"""
Custom key bindings
===================

Display one 4-D image layer using the ``add_image`` API

.. tags:: gui
"""

from skimage import data

import napari

blobs = data.binary_blobs(
    length=128, blob_size_fraction=0.05, n_dim=2, volume_fraction=0.25
).astype(float)

viewer = napari.Viewer()
layer = viewer.add_image(blobs, name='blobs')


@viewer.bind_key('a')
def accept_image(viewer):
    msg = 'this is a good image'
    viewer.status = msg
    print(msg)
    set_layer_data(viewer)


@viewer.bind_key('r')
def reject_image(viewer):
    msg = 'this is a bad image'
    viewer.status = msg
    print(msg)
    set_layer_data(viewer)


def set_layer_data(viewer):
    blobs = data.binary_blobs(
        length=128, blob_size_fraction=0.05, n_dim=2, volume_fraction=0.25
    ).astype(float)
    viewer.layers[0].data = blobs


@napari.Viewer.bind_key('w')
def hello(viewer):
    # on press
    viewer.status = 'hello world!'

    yield

    # on release
    viewer.status = 'goodbye world :('


# change viewer title
viewer.title = 'quality control images'

if __name__ == '__main__':
    napari.run()



from typing import TYPE_CHECKING

from napari._qt.layer_controls.qt_layer_controls_base import QtLayerControls
from napari._qt.layer_controls.widgets._tracks import (
    QtColormapComboBoxControl,
    QtColorPropertiesComboBoxControl,
    QtGraphCheckBoxControl,
    QtHeadLengthSliderControl,
    QtHideCompletedTracksCheckBoxControl,
    QtIdCheckBoxControl,
    QtTailDisplayCheckBoxControl,
    QtTailLengthSliderControl,
    QtTailWidthSliderControl,
)
from napari.layers.base._base_constants import Mode

if TYPE_CHECKING:
    import napari.layers


class QtTracksControls(QtLayerControls):
    """Qt view and controls for the Tracks layer.

    Parameters
    ----------
    layer : napari.layers.Tracks
        An instance of a Tracks layer.

    Attributes
    ----------
    _color_properties_combobox_control : napari._qt.layer_controls.widgets._tracks.QtColorPropertiesComboBoxControl
        Widget that wraps a comboBox controlling the layer color properties.
    _colormap_combobox_control : napari._qt.layer_controls.widgets._tracks.QtColormapComboBoxControl
        Widget that wraps a comboBox controlling current colormap of the layer.
    _graph_checkbox_control : napari._qt.layer_controls.widgets._tracks.QtGraphCheckBoxControl
        Checkbox controlling if graph of the layer should be shown.
    _head_length_slider_control : napari._qt.layer_controls.widgets._tracks.QtHeadLengthSliderControl
        Widget that wraps a slider controlling head length of the layer.
    _id_checkbox_control : napari._qt.layer_controls.widgets._tracks.QtIdCheckBoxControl
        Widget that wraps a checkbox controlling if id of the layer should be shown.
    _tail_display_checkbox_control : napari._qt.layer_controls.widgets._tracks.QtTailDisplayCheckBoxControl
        Widget that wraps a checkbox controlling if tails of the layer should be shown.
    _tail_length_slider_control : napari._qt.layer_controls.widgets._tracks.QtTailLengthSliderControl
        Widget that wraps a slider controlling tail length of the layer.
    _tail_width_slider_control : napari._qt.layer_controls.widgets._tracks.QtTailWidthSliderControl
        Widget that wraps a slider controlling tail width of the layer.
    _hide_completed_tracks_checkbox_control : napari._qt.layer_controls.widgets._tracks.QtHideCompletedTracksCheckBoxControl
        Widget that wraps a checkbox controlling if completed tracks of the layer should be hidden.
    """

    layer: 'napari.layers.Tracks'
    MODE = Mode
    PAN_ZOOM_ACTION_NAME = 'activate_tracks_pan_zoom_mode'
    TRANSFORM_ACTION_NAME = 'activate_tracks_transform_mode'

    def __init__(self, layer) -> None:
        super().__init__(layer)
        # Setup widgets controls
        self._color_properties_combobox_control = (
            QtColorPropertiesComboBoxControl(self, layer)
        )
        self._add_widget_controls(self._color_properties_combobox_control)
        self._colormap_combobox_control = QtColormapComboBoxControl(
            self, layer
        )
        self._add_widget_controls(self._colormap_combobox_control)
        self._tail_width_slider_control = QtTailWidthSliderControl(self, layer)
        self._add_widget_controls(self._tail_width_slider_control)
        self._tail_length_slider_control = QtTailLengthSliderControl(
            self, layer
        )
        self._add_widget_controls(self._tail_length_slider_control)
        self._head_length_slider_control = QtHeadLengthSliderControl(
            self, layer
        )
        self._add_widget_controls(self._head_length_slider_control)
        self._tail_display_checkbox_control = QtTailDisplayCheckBoxControl(
            self, layer
        )
        self._add_widget_controls(self._tail_display_checkbox_control)
        self._id_checkbox_control = QtIdCheckBoxControl(self, layer)
        self._add_widget_controls(self._id_checkbox_control)
        self._graph_checkbox_control = QtGraphCheckBoxControl(self, layer)
        self._add_widget_controls(self._graph_checkbox_control)
        self._hide_completed_tracks_checkbox_control = (
            QtHideCompletedTracksCheckBoxControl(self, layer)
        )
        self._add_widget_controls(self._hide_completed_tracks_checkbox_control)



"""
Colorbars and auto-tiling of overlays in grid mode
==================================================

Demonstrates the colorbar overlay for each layer, as well as the auto-tiling of
overlays. In addition, uses the `gridded` and `position` property of the scale
bar to show that overlays can also be auto-tiled per grid cell. This grid approach
also demonstrates that colorbars are linked to each layer in the grid, and by
toggling the grid mode off, the auto-tiling behavior can be seen.

.. tags:: visualization-basic
"""
import napari

viewer = napari.Viewer()
layers = viewer.open_sample('napari', 'lily')

# enable color bars, note that the colorbars will also tile according to
# each layer in the grid
for layer in layers:
    layer.colorbar.visible = True

# set the scale bar to gridded mode so it appears in each grid box
# have the position overlap with the default colorbar position
viewer.scale_bar.visible = True
viewer.scale_bar.position = 'top_right'
viewer.scale_bar.gridded = True

# enable grid with stride 2 to get layers split two-by-two
viewer.grid.enabled = True
viewer.grid.stride = 2

if __name__ == '__main__':
    napari.run()



"""
Labels 2D
=========

Display a labels layer above of an image layer using the ``add_labels`` and
``add_image`` APIs

.. tags:: visualization-basic
"""

from skimage import data
from skimage.color import rgb2gray
from skimage.segmentation import slic

import napari

astro = data.astronaut()

# initialise viewer with astro image
viewer = napari.Viewer()
layer = viewer.add_image(rgb2gray(astro), name='astronaut', rgb=False)

# add the labels
# we add 1 because SLIC returns labels from 0, which we consider background
labels = slic(astro, channel_axis=-1, compactness=20) + 1
label_layer = viewer.add_labels(labels, name='segmentation')

# Set the labels layer mode to picker with a string
label_layer.mode = 'PICK'
print(f'The color of label 5 is {label_layer.get_color(5)}')

if __name__ == '__main__':
    napari.run()



"""
Tracks 3D with graph
====================

.. tags:: visualization-advanced
"""

import numpy as np

import napari


def _circle(r, theta):
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return x, y


def tracks_3d_merge_split():
    """Create tracks with splitting and merging."""

    timestamps = np.arange(300)

    def _trajectory(t, r, track_id):
        theta = t * 0.1
        x, y = _circle(r, theta)
        z = np.zeros(x.shape)
        tid = np.ones(x.shape) * track_id
        return np.stack([tid, t, z, y, x], axis=1)

    trackA = _trajectory(timestamps[:100], 30.0, 0)
    trackB = _trajectory(timestamps[100:200], 10.0, 1)
    trackC = _trajectory(timestamps[100:200], 50.0, 2)
    trackD = _trajectory(timestamps[200:], 30.0, 3)

    data = [trackA, trackB, trackC, trackD]
    tracks = np.concatenate(data, axis=0)
    tracks[:, 2:] += 50.0  # centre the track at (50, 50, 50)

    graph = {1: 0, 2: [0], 3: [1, 2]}

    features = {'time': tracks[:, 1]}

    return tracks, features, graph


tracks, features, graph = tracks_3d_merge_split()
vertices = tracks[:, 1:]

viewer = napari.Viewer(ndisplay=3)
viewer.add_points(vertices, size=1, name='points', opacity=0.3)
viewer.add_tracks(tracks, features=features, graph=graph, name='tracks')

if __name__ == '__main__':
    napari.run()



"""
Surface with texture and vertex_colors
======================================

Display a 3D surface with both texture and color maps.

This example demonstrates how surfaces may be colored by:
    * setting `vertex_values`, which colors the surface with the selected
      `colormap`
    * setting `vertex_colors`, which replaces/overrides any color from
      `vertex_values`
    * setting both `texture` and `texcoords`, which blends a the value from
      a texture (image) with the underlying color from `vertex_values` or
      `vertex_colors`. Blending is achieved by multiplying the texture color by
      the underlying color - an underlying value of "white" will result in the
      unaltered texture color.

.. tags:: visualization-nD
"""

import numpy as np
from vispy.io import imread, load_data_file, read_mesh

import napari

# load the model and texture
mesh_path = load_data_file('spot/spot.obj.gz')
vertices, faces, _normals, texcoords = read_mesh(mesh_path)
n = len(vertices)
texture_path = load_data_file('spot/spot.png')
texture = imread(texture_path)

flat_spot = napari.layers.Surface(
    (vertices, faces),
    translate=(1, 0, 0),
    texture=texture,
    texcoords=texcoords,
    shading='flat',
    name='texture only',
)

np.random.seed(0)
plasma_spot = napari.layers.Surface(
    (vertices, faces, np.random.random((3, 3, n))),
    texture=texture,
    texcoords=texcoords,
    colormap='plasma',
    shading='smooth',
    name='vertex_values and texture',
)

rainbow_spot = napari.layers.Surface(
    (vertices, faces),
    translate=(-1, 0, 0),
    texture=texture,
    texcoords=texcoords,
    # the vertices are _roughly_ in [-1, 1] for this model and RGB values just
    # get clipped to [0, 1], adding 0.5 brightens it up a little :)
    vertex_colors=vertices + 0.5,
    shading='none',
    name='vertex_colors and texture',
)

# create the viewer and window
viewer = napari.Viewer(ndisplay=3)
viewer.add_layer(flat_spot)
viewer.add_layer(plasma_spot)
viewer.add_layer(rainbow_spot)

viewer.camera.center = (0.0, 0.0, 0.0)
viewer.camera.angles = (25.0, -50.0, -125.0)
viewer.camera.zoom = 150


if __name__ == '__main__':
    napari.run()



"""
Add grayscale image
===================

Display one grayscale image using the add_image API.

.. tags:: visualization-basic
"""

import numpy as np
from skimage import data

import napari

# simulating a grayscale image here for testing contrast limits adjustments
image = data.astronaut().mean(-1) * 100 + 100
image += np.random.rand(*image.shape) * 3000
viewer = napari.Viewer()
layer = viewer.add_image(image.astype(np.uint16))

if __name__ == '__main__':
    napari.run()



"""
Multithreading two-way
======================

.. tags:: interactivity
"""
import time

import numpy as np
from qtpy.QtWidgets import (
    QGridLayout,
    QLabel,
    QProgressBar,
    QPushButton,
    QWidget,
)

import napari
from napari.qt.threading import thread_worker


@thread_worker
def two_way_communication_with_args(start, end):
    """Both sends and receives values to & from the main thread.

    Accepts arguments, puts them on the worker object.
    Receives values from main thread with ``incoming = yield``
    Optionally returns a value at the end
    """

    # do computationally intensive work here
    i = start
    while i < end:
        i += 1
        time.sleep(0.1)
        # incoming receives values from the main thread
        # while yielding sends values back to the main thread
        incoming = yield i
        i = incoming if incoming is not None else i

    # do optional teardown here
    return 'done'


class Controller(QWidget):
    def __init__(self) -> None:
        super().__init__()

        layout = QGridLayout()
        self.setLayout(layout)
        self.status = QLabel('Click "Start"', self)
        self.play_btn = QPushButton('Start', self)
        self.abort_btn = QPushButton('Abort!', self)
        self.reset_btn = QPushButton('Reset', self)
        self.progress_bar = QProgressBar()

        layout.addWidget(self.play_btn, 0, 0)
        layout.addWidget(self.reset_btn, 0, 1)
        layout.addWidget(self.abort_btn, 0, 2)
        layout.addWidget(self.status, 0, 3)
        layout.setColumnStretch(3, 1)
        layout.addWidget(self.progress_bar, 1, 0, 1, 4)


def create_connected_widget():
    """Builds a widget that can control a function in another thread."""
    w = Controller()
    steps = 40

    # the decorated function now returns a GeneratorWorker object, and the
    # Qthread in which it's running.
    # (optionally pass start=False to prevent immediate running)
    worker = two_way_communication_with_args(0, steps)

    w.play_btn.clicked.connect(worker.start)

    # it provides signals like {started, yielded, returned, errored, finished}
    worker.returned.connect(lambda x: w.status.setText(f'worker returned {x}'))
    worker.errored.connect(lambda x: w.status.setText(f'worker errored {x}'))
    worker.started.connect(lambda: w.status.setText('worker started...'))
    worker.aborted.connect(lambda: w.status.setText('worker aborted'))

    # send values into the function (like generator.send) using worker.send
    # abort thread with worker.abort()
    w.abort_btn.clicked.connect(lambda: worker.quit())

    def on_reset_button_pressed():
        # we want to avoid sending into a unstarted worker
        if worker.is_running:
            worker.send(0)

    def on_yield(x):
        # Receive events and update widget progress
        w.progress_bar.setValue(100 * x // steps)
        w.status.setText(f'worker yielded {x}')

    def on_start():
        def handle_pause():
            worker.toggle_pause()
            w.play_btn.setText('Pause' if worker.is_paused else 'Continue')

        w.play_btn.clicked.disconnect(worker.start)
        w.play_btn.setText('Pause')
        w.play_btn.clicked.connect(handle_pause)

    def on_finish():
        w.play_btn.setDisabled(True)
        w.reset_btn.setDisabled(True)
        w.abort_btn.setDisabled(True)
        w.play_btn.setText('Done')

    w.reset_btn.clicked.connect(on_reset_button_pressed)
    worker.yielded.connect(on_yield)
    worker.started.connect(on_start)
    worker.finished.connect(on_finish)
    return w


if __name__ == '__main__':

    viewer = napari.Viewer()
    layer = viewer.add_image(np.random.rand(512, 512))
    w = create_connected_widget()
    viewer.window.add_dock_widget(w)

    napari.run()



"""
Add labels with features
========================

Display a labels layer with various features

.. tags:: layers, analysis, features-table
"""


import numpy as np
from skimage import data
from skimage.filters import threshold_otsu
from skimage.measure import label
from skimage.morphology import closing, remove_small_objects, square
from skimage.segmentation import clear_border

import napari

image = data.coins()[50:-50, 50:-50]

# apply threshold
thresh = threshold_otsu(image)
bw = closing(image > thresh, square(4))

# remove artifacts connected to image border
cleared = remove_small_objects(clear_border(bw), 20)

# label image regions
label_image = label(cleared)

# initialise viewer with coins image
viewer = napari.Viewer()
layer = viewer.add_image(image, name='coins', rgb=False)

# get the size of each coin (first element is background area)
label_areas = np.bincount(label_image.ravel())[1:]

# split coins into small or large
size_range = max(label_areas) - min(label_areas)
small_threshold = min(label_areas) + (size_range / 2)
coin_sizes = np.where(label_areas > small_threshold, 'large', 'small')

label_features = {
    'row': ['none']
    + ['top'] * 4
    + ['bottom'] * 4,  # background is row: none
    'size': ['none', *coin_sizes],  # background is size: none
}

colors = {1: 'white', 2: 'blue', 3: 'green', 4: 'red', 5: 'yellow',
          None: 'magenta'}
# Here we provide a dict with color mappings for a subset of labels;
# when passed to `add_labels`, using the `colormap` kwarg, it will be
# internally converted to a `napari.utils.colormaps.DirectLabelColormap`
# Note: we also provide a default color (`None` key) which will be used
# by all other labels

# add the labels
label_layer = viewer.add_labels(
    label_image,
    name='segmentation',
    features=label_features,
    colormap=colors,
)

if __name__ == '__main__':
    napari.run()



"""
To screenshot
=============

Display a variety of layer types in the napari viewer and take a screenshot of the viewer canvas with `viewer.screenshot()`.
The screenshot is then added back as an image layer.

Screenshots include all visible layers, bounded by the extent of the canvas, and is functional for 2D and 3D views.
To capture the extent of all data in 2D view, see `viewer.export_figure()`: :ref:`sphx_glr_gallery_export_figure.py` and :ref:`sphx_glr_gallery_screenshot_and_export_figure.py`.

This example code demonstrates screenshot shortcuts that do not include the viewer (e.g. `File` -> `Copy Screenshot to Clipboard`).
To include the napari viewer in the screenshot, use `viewer.screenshot(canvas_only=False)` or e.g. `File` -> `Copy Screenshot with Viewer to Clipboard`).

.. tags:: visualization-advanced
"""

import numpy as np
from skimage import data
from vispy.color import Colormap

import napari

# create the viewer and window
viewer = napari.Viewer()

# add the image
img_layer = viewer.add_image(data.camera(), name='photographer')
img_layer.colormap = 'gray'

# create a list of polygons
polygons = [
    np.array([[11, 13], [111, 113], [22, 246]]),
    np.array(
        [
            [505, 60],
            [402, 71],
            [383, 42],
            [251, 95],
            [212, 59],
            [131, 137],
            [126, 187],
            [191, 204],
            [171, 248],
            [211, 260],
            [273, 243],
            [264, 225],
            [430, 173],
            [512, 160],
        ]
    ),
    np.array(
        [
            [310, 382],
            [229, 381],
            [209, 401],
            [221, 411],
            [258, 411],
            [300, 412],
            [306, 435],
            [268, 434],
            [265, 454],
            [298, 461],
            [307, 461],
            [307, 507],
            [349, 510],
            [352, 369],
            [330, 366],
            [330, 366],
        ]
    ),
]

# add polygons
layer = viewer.add_shapes(
    polygons,
    shape_type='polygon',
    edge_width=1,
    edge_color='coral',
    face_color='royalblue',
    name='shapes',
)

# change some attributes of the layer
layer.selected_data = set(range(layer.nshapes))
layer.current_edge_width = 5
layer.opacity = 0.75
layer.selected_data = set()

# add an ellipse to the layer
ellipse = np.array([[59, 222], [110, 289], [170, 243], [119, 176]])
layer.add(
    ellipse,
    shape_type='ellipse',
    edge_width=5,
    edge_color='coral',
    face_color='purple',
)

masks = layer.to_masks([512, 512])
masks_layer = viewer.add_image(masks.astype(float), name='masks')
masks_layer.opacity = 0.7
masks_layer.colormap = Colormap([[0.0, 0.0, 0.0, 0.0], [1.0, 0.0, 0.0, 1.0]])

labels = layer.to_labels([512, 512])
labels_layer = viewer.add_labels(labels, name='labels')

points = np.array([[100, 100], [200, 200], [333, 111]])
size = np.array([10, 20, 20])
viewer.add_points(points, size=size)

# sample vector coord-like data
n = 100
pos = np.zeros((n, 2, 2), dtype=np.float32)
phi_space = np.linspace(0, 4 * np.pi, n)
radius_space = np.linspace(0, 100, n)

# assign x-y position
pos[:, 0, 0] = radius_space * np.cos(phi_space) + 350
pos[:, 0, 1] = radius_space * np.sin(phi_space) + 256

# assign x-y projection
pos[:, 1, 0] = 2 * radius_space * np.cos(phi_space)
pos[:, 1, 1] = 2 * radius_space * np.sin(phi_space)

# add the vectors
layer = viewer.add_vectors(pos, edge_width=2)

# take screenshot
screenshot = viewer.screenshot()
# optionally for saving the exported screenshot: viewer.screenshot(path="screenshot.png")
viewer.add_image(screenshot, rgb=True, name='screenshot')

if __name__ == '__main__':
    napari.run()



# See "Writing benchmarks" in the asv docs for more information.
# https://asv.readthedocs.io/en/latest/writing_benchmarks.html
# or the napari documentation on benchmarking
# https://github.com/napari/napari/blob/main/docs/BENCHMARKS.md
import numpy as np

from napari.layers import Surface


class Surface2DSuite:
    """Benchmarks for the Surface layer with 2D data"""

    params = [2**i for i in range(4, 18, 2)]

    def setup(self, n):
        np.random.seed(0)
        self.data = (
            np.random.random((n, 2)),
            np.random.randint(n, size=(n, 3)),
            np.random.random(n),
        )
        self.layer = Surface(self.data)

    def time_create_layer(self, n):
        """Time to create an image layer."""
        Surface(self.data)

    def time_refresh(self, n):
        """Time to refresh view."""
        self.layer.refresh()

    def time_set_view_slice(self, n):
        """Time to set view slice."""
        if hasattr(self.layer, '_slicing_state'):
            self.layer._slicing_state._set_view_slice()
        else:
            # before https://github.com/napari/napari/pull/8254
            self.layer._set_view_slice()

    def time_update_thumbnail(self, n):
        """Time to update thumbnail."""
        self.layer._update_thumbnail()

    def time_get_value(self, n):
        """Time to get current value."""
        self.layer.get_value((0,) * 2)

    def mem_layer(self, n):
        """Memory used by layer."""
        return self.layer

    def mem_data(self, n):
        """Memory used by raw data."""
        return self.data


class Surface3DSuite:
    """Benchmarks for the Surface layer with 3D data."""

    params = [2**i for i in range(4, 18, 2)]

    def setup(self, n):
        np.random.seed(0)
        self.data = (
            np.random.random((n, 3)),
            np.random.randint(n, size=(n, 3)),
            np.random.random(n),
        )
        self.layer = Surface(self.data)

    def time_create_layer(self, n):
        """Time to create a layer."""
        Surface(self.data)

    def time_refresh(self, n):
        """Time to refresh view."""
        self.layer.refresh()

    def time_set_view_slice(self, n):
        """Time to set view slice."""
        if hasattr(self.layer, '_slicing_state'):
            self.layer._slicing_state._set_view_slice()
        else:
            # before https://github.com/napari/napari/pull/8254
            self.layer._set_view_slice()

    def time_update_thumbnail(self, n):
        """Time to update thumbnail."""
        self.layer._update_thumbnail()

    def time_get_value(self, n):
        """Time to get current value."""
        self.layer.get_value((0,) * 3)

    def mem_layer(self, n):
        """Memory used by layer."""
        return self.layer

    def mem_data(self, n):
        """Memory used by raw data."""
        return self.data


if __name__ == '__main__':
    from utils import run_benchmark

    run_benchmark()



import json
import logging
import pathlib
from argparse import ArgumentParser

import matplotlib.pyplot as plt

logging.basicConfig(
    format='%(levelname)s : %(asctime)s : %(message)s',
    level=logging.INFO,
)

parser = ArgumentParser(
    description='Plot the durations of a callable measured by perfmon.',
)
parser.add_argument(
    'config',
    help='The name of the sub-directory that contains the perfmon traces (e.g. slicing)',
)
parser.add_argument(
    'callable',
    help='The name of the callable to plot excluding the module (e.g. QtDimSliderWidget._value_changed).',
)
parser.add_argument(
    'baseline',
    default='baseline',
    help='The name added to output traces file for the baseline measurement.',
)
parser.add_argument(
    'test',
    default='test',
    help='The name added to output traces file for the test measurement.',
)
args = parser.parse_args()

logging.info(
    'Running compare_callable.py with the following arguments.\n{args_}',
    extra={'args_': args},
)

perfmon_dir = pathlib.Path(__file__).parent.resolve(strict=True)
config_dir = perfmon_dir / args.config


def _get_durations_ms(output_name: str) -> list[float]:
    file_path = str(config_dir / f'traces-{output_name}.json')
    with open(file_path) as traces_file:
        traces = json.load(traces_file)
    return [
        trace['dur'] / 1000
        for trace in traces
        if trace['name'] == args.callable
    ]


baseline_durations_ms = _get_durations_ms(args.baseline)
test_durations_ms = _get_durations_ms(args.test)

plt.violinplot(
    [baseline_durations_ms, test_durations_ms],
    vert=False,
)

plt.title(f'{args.config}: {args.callable} ({args.baseline} vs. {args.test})')
plt.xlabel('Duration (ms)')
plt.yticks([1, 2], [args.baseline, args.test])

plt.show()



# See "Writing benchmarks" in the asv docs for more information.
# https://asv.readthedocs.io/en/latest/writing_benchmarks.html
# or the napari documentation on benchmarking
# https://github.com/napari/napari/blob/main/docs/BENCHMARKS.md
import os

import numpy as np
from packaging.version import parse as parse_version
from qtpy.QtWidgets import QApplication

import napari

NAPARI_0_4_19 = parse_version(napari.__version__) <= parse_version('0.4.19')


class QtViewerViewVectorSuite:
    """Benchmarks for viewing vectors in the viewer."""

    params = [2**i for i in range(4, 18, 2)]

    if 'PR' in os.environ:
        skip_params = [(2**i,) for i in range(8, 18, 2)]

    def setup(self, n):
        _ = QApplication.instance() or QApplication([])
        np.random.seed(0)
        self.data = np.random.random((n, 2, 3))
        self.viewer = napari.Viewer()
        self.layer = self.viewer.add_vectors(self.data)
        if NAPARI_0_4_19:
            self.visual = self.viewer.window._qt_viewer.layer_to_visual[
                self.layer
            ]
        else:
            self.visual = self.viewer.window._qt_viewer.canvas.layer_to_visual[
                self.layer
            ]

    def teardown(self, n):
        self.viewer.window.close()

    def time_vectors_refresh(self, n):
        """Time to refresh a vector."""
        self.viewer.layers[0].refresh()

    def time_vectors_multi_refresh(self, n):
        """Time to refresh a vector multiple times."""
        self.viewer.layers[0].refresh()
        self.viewer.layers[0].refresh()
        self.viewer.layers[0].refresh()


if __name__ == '__main__':
    from utils import run_benchmark

    run_benchmark()



"""
nD image
========

Display one 4-D image layer using the :func:`add_image` API.

.. tags:: visualization-nD
"""

import numpy as np
from skimage import data

import napari

blobs = np.stack(
    [
        data.binary_blobs(
            length=128, blob_size_fraction=0.05, n_dim=3, volume_fraction=f
        )
        for f in np.linspace(0.05, 0.5, 10)
    ],
    axis=0,
)
viewer = napari.Viewer()
layer = viewer.add_image(blobs.astype(float))

if __name__ == '__main__':
    napari.run()
